import numpy as np
import melee


class ObservationSpace:
    def __init__(self):
        self.size = 18  # better way to set this?
        self.current_frame = 0
        self.done = False
        self.previous_gamestate = None
        self.current_gamestate = None

    def _reset(self):
        self.current_frame = 0
        self.done = False
        self.previous_gamestate = None
        self.current_gamestate = None

    def get_stocks(self):
        stocks = [self.current_gamestate.player[i].stock for i in list(self.current_gamestate.player.keys())]
        return np.array([stocks]).T #playersx1
    
    def get_actions(self):
        actions = [self.current_gamestate.player[i].action.value for i in list(self.current_gamestate.player.keys())]
        action_frames = [self.current_gamestate.player[i].action_frame for i in list(self.current_gamestate.player.keys())]
        hitstun_frames_left = [self.current_gamestate.player[i].hitstun_frames_left for i in list(self.current_gamestate.player.keys())]
        
        return np.array([actions, action_frames, hitstun_frames_left]).T # playersx3
        
    def __call__(self, gamestate):
        """ pull out relevant info from gamestate """
        self.current_gamestate = gamestate
        self.current_frame +=1 
        total_reward = 0
        info = None
        
        actions = self.get_actions()
        stocks = self.get_stocks()

        # if 0 <= actions[0][0] <= 10:  # DEAD_DOWN, DEAD_LEFT ... 
        #     info = "dead"
        
        # self.done = not (bool(stocks[1][0]) or bool(stocks[1][1]))
        if self.current_frame > 85 and not self.done:
            # reward/penalize based on delta damage/stocks
            # +- 1 for stocks taken/lost
            # total_reward += (stocks[0][1])
            # total_reward -= (stocks[0][0])

            # # if stocks change, don't reward for going back to 0 damage
            # # assumption: you can't lose more than one stock on a given frame
            # total_reward -= (percents[0][0] * (1^stocks[0][0])) * .01
            # total_reward += (percents[0][1] * (1^stocks[0][1])) * .01
            total_reward = 0

        if self.current_gamestate is not None:
            self.previous_gamestate = self.current_gamestate

        
        observation = np.concatenate((actions, stocks), axis=1)

        return observation, total_reward, self.done, info


class ActionSpace:
    def __init__(self):
        mid = np.sqrt(2)/2

        self.stick_space_reduced = np.array([[0.0, 0.0], # no op
                                             [0.0, 1.0],
                                             [mid, mid],
                                             [1.0, 0.0],
                                             [mid, -mid],
                                             [0.0, -1.],
                                             [-mid, -mid],
                                             [-1., 0.0],
                                             [-mid, mid]])

        self.button_space_reduced = np.array([0., 1., 2., 3., 4.])

        # Action space size is total number of possible actions. In this case,
        #    is is all possible main stick positions * all c-stick positions *
        #    all the buttons. A normal controller has ~51040 possible main stick 
        #    positions. Each trigger has 255 positions. The c-stick can be 
        #    reduced to ~5 positions. Finally, if all buttons can be pressed
        #    in any combination, that results in 32 combinations. Not including
        #    the dpad or start button, that is 51040 * 5 * 255 * 2 * 32 which 
        #    is a staggering 4.165 billion possible control states. 

        # Given this, it is reasonable to reduce this. In the above class, the 
        #    main stick has been reduced to the 8 cardinal positions plus the 
        #    center (no-op). Only A, B, Z, and R are used, as these correspond
        #    to major in-game functions (attack, special, grab, shield). Every
        #    action can theoretically be performed with just these buttons. A 
        #    final "button" is added for no-op. 
        #
        #    Action space = 9 * 5 = 45 possible actions. 
        self.action_space = np.zeros((self.stick_space_reduced.shape[0] * self.button_space_reduced.shape[0], 3))

        for button in self.button_space_reduced:
            self.action_space[int(button)*9:(int(button)+1)*9, :2] = self.stick_space_reduced
            self.action_space[int(button)*9:(int(button)+1)*9, 2] = button

        # self.action_space will look like this, where the first two columns
        #   represent the control stick's position, and the final column is the 
        #   currently depressed button. 

        # ACT  Left/Right    Up/Down      Button
        # ---  ------        ------       ------
        # 0   [ 0.        ,  0.        ,  0. (NO-OP)] Center  ---
        # 1   [ 0.        ,  1.        ,  0.        ] Up         |
        # 2   [ 0.70710678,  0.70710678,  0.        ] Up/Right   |
        # 3   [ 1.        ,  0.        ,  0.        ] Right      |
        # 4   [ 0.70710678, -0.70710678,  0.        ] Down/Right |- these repeat
        # 5   [ 0.        , -1.        ,  0.        ] Down       |
        # 6   [-0.70710678, -0.70710678,  0.        ] Down/Left  |
        # 7   [-1.        ,  0.        ,  0.        ] Left       |
        # 8   [-0.70710678,  0.70710678,  0.        ] Up/Left  ---
        # 9   [ 0.        ,  0.        ,  1. (A)    ] 
        # 10  [ 0.        ,  1.        ,  1.        ] 
        # 11  [ 0.70710678,  0.70710678,  1.        ]
        # 12  [ 1.        ,  0.        ,  1.        ]
        # 13  [ 0.70710678, -0.70710678,  1.        ]
        # 14  [ 0.        , -1.        ,  1.        ]
        # 15  [-0.70710678, -0.70710678,  1.        ]
        # 16  [-1.        ,  0.        ,  1.        ]
        # 17  [-0.70710678,  0.70710678,  1.        ]
        # 18  [ 0.        ,  0.        ,  2. (B)    ] 
        # 19  [ 0.        ,  1.        ,  2.        ]
        # 20  [ 0.70710678,  0.70710678,  2.        ]
        # 21  [ 1.        ,  0.        ,  2.        ]
        # 22  [ 0.70710678, -0.70710678,  2.        ]
        # 23  [ 0.        , -1.        ,  2.        ]
        # 24  [-0.70710678, -0.70710678,  2.        ]
        # 25  [-1.        ,  0.        ,  2.        ]
        # 26  [-0.70710678,  0.70710678,  2.        ]
        # 27  [ 0.        ,  0.        ,  3. (Z)    ] 
        # 28  [ 0.        ,  1.        ,  3.        ]
        # 29  [ 0.70710678,  0.70710678,  3.        ]
        # 30  [ 1.        ,  0.        ,  3.        ]
        # 31  [ 0.70710678, -0.70710678,  3.        ]
        # 32  [ 0.        , -1.        ,  3.        ]
        # 33  [-0.70710678, -0.70710678,  3.        ]
        # 34  [-1.        ,  0.        ,  3.        ]
        # 35  [-0.70710678,  0.70710678,  3.        ] 
        # 36  [ 0.        ,  0.        ,  4. (R)    ] 
        # 37  [ 0.        ,  1.        ,  4.        ]
        # 38  [ 0.70710678,  0.70710678,  4.        ]
        # 39  [ 1.        ,  0.        ,  4.        ]
        # 40  [ 0.70710678, -0.70710678,  4.        ]
        # 41  [ 0.        , -1.        ,  4.        ]
        # 42  [-0.70710678, -0.70710678,  4.        ]
        # 43  [-1.        ,  0.        ,  4.        ]
        # 45  [-0.70710678,  0.70710678,  4.        ]

        self.size = self.action_space.shape[0]

    def sample(self):
        return np.random.choice(self.size)

    def __call__(self, action):
        if action > self.size - 1:
            exit("Error: invalid action!")

        return ControlState(self.action_space[action])


class ControlState:
    def __init__(self, state):
        self.state = state
        self.buttons = [
            False,
            melee.enums.Button.BUTTON_A,
            melee.enums.Button.BUTTON_B,
            melee.enums.Button.BUTTON_Z,
            melee.enums.Button.BUTTON_R]

    def execute(self, controller):
        controller.release_all()      
        if self.state[2]:             # only press button if not no-op
            if self.state[2] != 4.0:  # special case for r shoulder
                controller.press_button(self.buttons[int(self.state[2])]) 
            else:
                controller.press_shoulder(melee.enums.Button.BUTTON_R, 1)
        
        controller.tilt_analog_unit(melee.enums.Button.BUTTON_MAIN, 
                                    self.state[0], self.state[1])



from melee_env.dconfig import DolphinConfig
import melee
from melee import enums
import numpy as np
    
class MeleeEnv:
    def __init__(self, 
        iso_path,
        players,
        action_space,
        observation_space,
        fast_forward=False, 
        blocking_input=False,
        ai_starts_game=False):

        self.d = DolphinConfig()
        self.d.set_ff(fast_forward)

        self.iso_path = iso_path
        self.players = players

        self.action_space = action_space 
        self.observation_space = observation_space

        self.blocking_input = blocking_input
        self.ai_starts_game = ai_starts_game

        self.gamestate = None


    def start(self):
        self.console = melee.Console(
            path=str(self.d.slippi_bin_path),
            dolphin_home_path=str(self.d.slippi_home)+"/",
            blocking_input=self.blocking_input)  # broken in 0.21.0

        # Configure Dolphin for the correct controller setup, add controllers
        for i in range(len(self.players)):
            curr_player = self.players[i]

            if curr_player.agent_type == "HMN":
                self.d.set_controller_type(i+1, enums.ControllerType.GCN_ADAPTER)
                curr_player.controller = melee.Controller(console=self.console, port=i+1, type=melee.ControllerType.GCN_ADAPTER)
                curr_player.port = i+1
            elif curr_player.agent_type in ["AI", "CPU"]:
                self.d.set_controller_type(i+1, enums.ControllerType.STANDARD)
                curr_player.controller = melee.Controller(console=self.console, port=i+1)
                self.menu_control_agent = i
                curr_player.port = i+1 
            else:  # no player
                self.d.set_controller_type(i+1, enums.ControllerType.UNPLUGGED)
        
        self.console.run(iso_path=self.iso_path)
        self.console.connect()


        [player.controller.connect() for player in self.players if player is not None]

        self.gamestate = self.console.step()
 
    def setup(self, stage):
        while True:
            self.gamestate = self.console.step()
            if self.gamestate.menu_state is melee.Menu.CHARACTER_SELECT:
                for i in range(len(self.players)):
                    if self.players[i].agent_type == "AI":
                        melee.MenuHelper.choose_character(
                            character=self.players[i].character,
                            gamestate=self.gamestate,
                            controller=self.players[i].controller,
                            costume=i,
                            swag=False,
                            start=self.ai_starts_game)
                    if self.players[i].agent_type == "CPU":
                        melee.MenuHelper.choose_character(
                            character=self.players[i].character,
                            gamestate=self.gamestate,
                            controller=self.players[i].controller,
                            costume=i,
                            swag=False,
                            cpu_level=self.players[i].lvl,
                            start=False)  # HMN needed to start cpu-only game

            elif self.gamestate.menu_state is melee.Menu.STAGE_SELECT:
                melee.MenuHelper.choose_stage(
                    stage=stage,
                    gamestate=self.gamestate,
                    controller=self.players[self.menu_control_agent].controller)

            elif self.gamestate.menu_state in [melee.Menu.IN_GAME, melee.Menu.SUDDEN_DEATH]:
                if self.gamestate.frame < -83:
                    continue
                else:
                    return self.observation_space(self.gamestate)

            else:
                melee.MenuHelper.choose_versus_mode(self.gamestate, self.players[self.menu_control_agent].controller)


    def step(self):
        if self.gamestate.menu_state in [melee.Menu.IN_GAME, melee.Menu.SUDDEN_DEATH]:
            for i in range(len(self.players)):
                if self.players[i].agent_type == "AI":
                    # `action` is one of N possible actions an agent can take
                    #   as defined in action_space
                    controller_input = self.action_space(self.players[i].action)

                    # execute must be written to support any action from 
                    #   ActionSpace
                    controller_input.execute(self.players[i].controller)    

            self.gamestate = self.console.step()

            observation, reward, done, info = self.observation_space(self.gamestate)

            for i in range(len(self.players)):
                if self.players[i].defeated == True and len(observation) < len(self.players):
                    observation = np.insert(observation, i, np.full(observation.shape[1], -1), axis=0)

            return observation, reward, done, info
        else:
            return None, None, True, None

    def close(self):
        for t, c in self.controllers.items():
            c.disconnect()
        self.observation_space._reset()
        self.gamestate = None
        self.console.stop()
        time.sleep(2) 


import configparser
from pathlib import Path
import sys
import melee
import requests
from tqdm import tqdm
import os
import subprocess
import shutil
import time
import signal

class DolphinConfig:
    def __init__(self):

        # setup paths
        self.home = Path.home()
        self.platform = sys.platform
        
        self.module_path = Path(__file__).resolve().parents[0]  # module installation path
        self.install_data_path = self.module_path / "install_data"

        if self.platform == "win32":
            self.data_path = self.home / "AppData/Roaming"
        elif self.platform == "linux":
            self.data_path = self.home / ".local/share"
        elif self.platform == "darwin":
            self.data_path = self.home / "Library/Application Support"

        self.slippi_path = self.data_path / "melee-env" / "Slippi"

        self.slippi_replays_path = self.slippi_path / "replays"
        self.slippi_bin_path = self.slippi_path / "squashfs-root/usr/bin"
        self.squash_fs = self.slippi_path / "squashfs-root"
        self.slippi_home = self.slippi_path / "data"
        self.config_path = self.slippi_home / "Config/Dolphin.ini"

        self.slippi_gecko_ini_path = self.slippi_bin_path / "Sys/GameSettings/GALE01r2.ini"

        self.install_data_path = self.module_path / "install_data"

        # check that our local slippi is installed
        if not self.slippi_bin_path.exists():
            # assume this is a new installation?

            # need to download slippi (based on platform!)
            self.install_slippi(self.slippi_path)

            # download gecko codes for slippi
            self.apply_gecko_codes(self.slippi_path)

            # extra configuration steps
            self.configure_dolphin(self.slippi_path)

            print(f"Successfully downloaded, installed, and configured dolphin "
                f" in {self.slippi_path.parents[0]}")

        else:
            # check for updates? might be nice to preserve the current install
            # in case things break. 
            print(f"Found melee-env installation in {self.slippi_path.parents[0]}")


    def use_render_interface(self, interface="opengl"):
        """Edit config to use Vulkan instead of default OpenGL"""
        if interface not in ["vulkan", "opengl"]:
            raise ValueError("unsupported render interface, please select "
                "either 'vulkan' or 'opengl'")

        if interface == "vulkan":
            config = configparser.ConfigParser()
            config.readfp(open(self.config_path))
            config['Core']['gfxbackend'] = "Vulkan"
        else:
            config = configparser.ConfigParser()
            config.readfp(open(self.config_path))
            config['Core']['gfxbackend'] = ""

            with open(str(self.config_path), 'w') as outfile:
                config.write(outfile)

        return

    def set_ff(self, enable=True):
        """Edit GALE01r2.ini to enable fast forward. Useful for faster training"""
        with open(str(self.slippi_gecko_ini_path), 'r') as f:
            data = f.readlines()

        if "Fast Forward" not in data[18]:
            raise FileNotFoundError(f"Error: cannot locate Fast Forward Gecko code in "
                 "{self.slippi_gecko_ini_path}, please ensure it is located in this "
                 "file, and that it is on line 19!")

        if data[18][0] == "-" and enable: 
            data[18] = "$Optional: Fast Forward\n"
            with open(str(self.slippi_gecko_ini_path), 'w') as f:
                f.writelines(data)

        elif data[18][0] == "$" and not enable:
            data[18] = "-Optional: Fast Forward\n"
            with open(str(self.slippi_gecko_ini_path), 'w') as f:
                f.writelines(data)

        return

    def set_controller_type(self, port, controller_type):
        if  not (1 <= int(port) <= 4):
            raise ValueError(f"Port must be 1, 2, 3, or 4. Received value {port}")

        port = int(port)

        if controller_type not in [e for e in melee.enums.ControllerType]:
            raise ValueError(f"Controller type must be one of {{{[e.name+':'+e.value for e in melee.enums.ControllerType]}}}")

        config = configparser.ConfigParser()
        config.readfp(open(self.config_path))
        config['Core'][f'sidevice{str(port-1)}'] = f"{controller_type.value}"

        with open(str(self.config_path), 'w') as outfile:
            config.write(outfile)

        return


    def _download_file(self, url):
        local_filename = url.split('/')[-1]
        # NOTE the stream=True parameter below
        with requests.get(url, stream=True) as r:
            total_size_in_bytes= int(r.headers.get('content-length', 0))
            progress_bar = tqdm(
                total=total_size_in_bytes, 
                unit='iB', 
                unit_scale=True,
                desc=f"Downloading {local_filename}")

            r.raise_for_status()
            with open(local_filename, 'wb') as f:
                for chunk in r.iter_content(chunk_size=8192): 
                    progress_bar.update(len(chunk))
                    f.write(chunk)

        progress_bar.close()
        return Path(local_filename).resolve()

    def install_slippi(self, install_path):
        if self.platform == "linux":
            target_url = "https://github.com/project-slippi/Ishiiruka/releases/latest/download/Slippi_Online-x86_64.AppImage"

        elif self.platform == "win32":
            target_url = "https://github.com/project-slippi/Ishiiruka/releases/latest/download/Slippi_Online-x86_64.zip"
            raise NotImplementedError("Windows currently not supported at this time.")

        elif self.platform == "darwin":
            target_url = "https://github.com/project-slippi/Ishiiruka/releases/download/v2.3.1/FM-Slippi-2.3.1-Mac.dmg"
            raise NotImplementedError("OSX currently not supported at this time.")

        install_path.mkdir(parents=True, exist_ok=True)

        # Download latest version of slippi
        slippi_game_path = self._download_file(target_url)

        # move to our directory
        slippi_game_path.rename(install_path / slippi_game_path.name)
        slippi_game_path = install_path / slippi_game_path.name

        if self.platform == "linux":
            print("Dolphin will open and then close to generate files")
            # Set as executable
            slippi_game_path.chmod(509)

            os.chdir(str(slippi_game_path.parents[0]))
            cmd_extract = f"{slippi_game_path} --appimage-extract"

            # Extract 
            process = subprocess.Popen(cmd_extract.split(), stdout=subprocess.PIPE)
            output, error = process.communicate()

            apprun_path = self.squash_fs / "AppRun"
            print(f"Running: {str(apprun_path)+f' -u {str(self.slippi_home)}'}")
            process = subprocess.Popen(
                str(apprun_path)+f" -u {str(self.slippi_home)}", 
                stdout=subprocess.PIPE, 
                shell=True,
                preexec_fn=os.setsid)

            time.sleep(2)
            os.killpg(os.getpgid(process.pid), signal.SIGTERM) # send kill signal

            # Delete executable
            slippi_game_path.unlink()

        if self.platform == "win32":
            # unzip the archive, etc. See libmelee readme
            pass

    def apply_gecko_codes(self, install_path):
        # get most up-to-date codes:
        gale01r2_url = "https://raw.githubusercontent.com/altf4/slippi-ssbm-asm/libmelee/Output/Netplay/GALE01r2.ini"
        gale01r2_path = self._download_file(gale01r2_url)

        # Rename the old ini file
        self.slippi_gecko_ini_path.rename(self.slippi_gecko_ini_path.parents[0] / "GALE01r2.ini.old")

        # Copy the required codes to the old location
        dest = shutil.copy(str(gale01r2_path), str(self.slippi_gecko_ini_path))

        # Add the fastforward code at the end of this file:
        fast_forward = open(self.install_data_path / "fast_forward").read()
        with open(str(self.slippi_gecko_ini_path), 'a') as f:
            f.write("\n")
            f.write(fast_forward)

        # Add the fast forward code to the [Gecko_Enabled] section
        with open(str(self.slippi_gecko_ini_path), 'r') as f:
            data = f.readlines()

        # Surely the below can be done more cleanly with configparser
        # turn off recommended: apply delay to all in-game scenes
        if data[13] == "$Recommended: Apply Delay to all In-Game Scenes\n":
            data[13] = "-" + data[13][1:]  # turn this off

        # check that line 15 is empty before we start adding stuff
        if data[14] != "\n":
            raise FileNotFoundError("Something has gone wrong... check that {self.slippi_gecko_ini_path} exists.")

        data.insert(14, "-Optional: Fast Forward\n")
        data.insert(14, "-Optional: Flash Red on Failed L-Cancel\n")
        data.insert(14, "$Optional: Center Align 2P HUD\n")
        data.insert(14, "$Optional: Disable Screen Shake\n")
        data.insert(14, "-Optional: Widescreen 16:9\n")

        with open(str(self.slippi_gecko_ini_path), 'w') as f:
            f.writelines(data)

        # cleanup
        gale01r2_path.unlink()

    def configure_dolphin(self, install_path):

        if not self.config_path.exists():
            raise FileNotFoundError(f"No Slippi Config. Run Slippi Online once to generate these files"
                 f"Then confirm that files exist in {self.config_path}")
        

        config = configparser.ConfigParser()
        config.read_file(open(self.config_path))
        config['Core']['SlippiReplayDir'] = str(self.slippi_replays_path)
        config['Core']['SlippiReplayMonthFolders'] = "True"

        with open(str(self.config_path), 'w') as outfile:
            config.write(outfile)
         
        # Link this config to the extracted dolphin
        user_path = self.slippi_bin_path / "User"
        user_path.mkdir(parents=True, exist_ok=True)

        config_path = user_path / "Config"

        if self.platform == "linux":
            config_path.symlink_to(self.config_path.parents[0])

        # Move GCPadNew.ini to this location
        shutil.copy(
            str(self.install_data_path / "GCPadNew.ini"), 
            str(self.config_path.parents[0] / "GCPadNew.ini")
            )

if __name__=="__main__":
    # test installation
    d = DolphinConfig()


from abc import ABC, abstractmethod
from melee import enums

class Agent(ABC):
    def __init__(self):
        self.agent_type = None
        self.controller = None
        self.port = None
        self.action = None
        self.defeated = False

    def _is_defeated(self, observation):
        if self.defeated:
            return True
        
        elif observation[3] == 0 and not self.defeated:
            self.defeated = True
            self.action = 0
            print(f"{self} got beat")
            return True

        else:
            return False

    @abstractmethod
    def act(self):
        pass

class AgentChooseCharacter(Agent):
    def __init__(self, character):
        super().__init__()
        self.character = character
        

class Human(Agent):
    def __init__(self):
        super().__init__()
        self.agent_type = "HMN"
        
    def act():
        pass


class CPU(AgentChooseCharacter):
    def __init__(self, character, lvl):
        super().__init__(character)
        self.agent_type = "CPU"
        if not 1 <= lvl <= 9:
            raise ValueError(f"CPU Level must be 1-9. Got {lvl}")
        self.lvl = lvl
        
    def act():
        pass


class Random(AgentChooseCharacter):
    def __init__(self, character):
        super().__init__(character)
        self.agent_type = "AI"
           
    def act(self, observation, action_space):
        if not self._is_defeated(observation):
            action = action_space.sample()
            self.action = action


class Shine(Agent):
    # refer to action_space.py
    def __init__(self):
        super().__init__()
        self.agent_type = "AI"
        self.character = enums.Character.FOX
        
    def act(self, observation, action_space):
        # lib/melee reports defeated state for 60 frames, then completely drops
        #   the player from reporting
        if not self._is_defeated(observation):
            action = 0  # none 

            if (observation[0] == enums.Action.STANDING.value or 
                observation[0] == enums.Action.CROUCH_START.value):
                action = 5  # crouch

            if observation[0] == enums.Action.CROUCHING.value:
                action = 23  # down-B (shine)

            if observation[0] == enums.Action.KNEE_BEND.value and observation[1] == 3:
                    action = 23  # shine again on frame 3 of knee bend.

            if (observation[0] == enums.Action.DOWN_B_GROUND.value or 
                observation[0] == enums.Action.DOWN_B_GROUND_START.value):
                action = 10  # tap jump

            if observation[2] > 0:
                action = 0  # don't do crazy things while in hitstun
            
            self.action = action
